# Copyright (c) 2004 - 2008 16 Systems, LLC
# Copyright (c) 2009 Virginia Polytechnic Institute and State University
# All rights reserved.
#
# License & Redistribution Information
# http://www.security.vt.edu/Find_SSNs/find_ssns_license.html


import re
import os
import os.path
import time
import version
import zipfile


# This module should be broken into 2 or 3 smaller modules.
# (hunt, validate, report) when time permits.
# It has grown too large and is becoming difficult to change.


def go(cs, ag, ms, search_path, report_type, skipped_numbers, target_file_list, path_to_results, reg_exp, search_for_value, mode):

    if mode == 'gui':
        import wx

    # Delimiters to be removed.
    bad = re.compile(r'[-\s.|]')

    tt_time = time.strftime("%a_%b_%d_%Y_%H_%M_%S_%p", time.localtime())

    # The list of suspect files:
    suspect = []

    # Increase threshold to reduce false positives.
    # Set dthreshold (default threshold) to equal threshold
    threshold = 1
    dthreshold = threshold

    if os.path.isfile('threshold.txt'):
        fp = open('threshold.txt')
        # Change threshold to user specified number.
        try:
            threshold = int(fp.read().strip())
        # 'threshold.txt' contains a non-numeric, fallback to dthreshold.
        except ValueError:
            threshold = dthreshold
        fp.close()

    #print threshold

    results = open(os.path.join(path_to_results, 'Find_SSNs.%s' %report_type), 'a')

    # Validate contains the two functions valid_ccn and valid_ssn
    # Could be re-written in C++ and used with boost.python
    # to speed-up performance. 
    def validate(numbers, filename):

        details = open(os.path.join(path_to_results, 'Find_SSNs.txt'), 'a')

        #print numbers, filename

        def valid_ccn(ccn):


            #print ccn

            ############################## Start

            # Author David Shaw (Luhn Algorithm)
            # Source http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/172845
            # Modified slightly by 16 Systems, LLC

            sumit = 0
            num_digits = len(ccn)
            oddeven = num_digits & 1

            for count in range(0, num_digits):
                digit = int(ccn[count])

                if not (( count & 1 ) ^ oddeven ):
                    digit = digit * 2
                if digit > 9:
                    digit = digit - 9

                sumit = sumit + digit

            ############################### End

            if (sumit % 10) == 0:
                # Any card with 1 digit prefix
                if ccn[0] in cs.keys():
                    for k,v in cs.iteritems():
                        if ccn[0] == k and len(ccn) in v[0]:
                            return (True, True)
                        else:
                            return (False, False)

                # Any card with 2 digit prefix
                elif ccn[:2] in cs.keys():
                    for k,v in cs.iteritems():
                        if ccn[:2] == k and len(ccn) == v[0]:
                            return (True, True)
                        else:
                            return (False, False)

                # Any card with 3 digit prefix
                elif ccn[:3] in cs.keys():
                    for k,v in cs.iteritems():
                        if ccn[:3] == k and len(ccn) == v[0]:
                            return (True, True)
                        else:
                            return (False, False)

                # Any card with 4 digit prefix
                elif ccn[:4] in cs.keys():
                    for k,v in cs.iteritems():
                        if ccn[:4] == k and len(ccn) == v[0]:
                            return (True, True)
                        else:
                            return (False, False)

                # Obviously not a CC we're looking for, but note that there are other cards out there.
                else:
                    #print ccn, "Not a credit card we're searching for."
                    return (False, False)
            else:
                return (False, False)

        # End of valid_ccn


        def valid_ssn(area_number, group_number, serial_number):

            if group_number == '00':
                return (False, False)

            if serial_number == '0000':
                return (False, False)

            odd_one = [1,3,5,7,9]
            even_two = range(10, 100, 2) # (10-98 even only)
            even_three = [2,4,6,8]
            odd_four = range(11, 100, 2)# (11-99 odd only)

            if area_number in ag.keys():
                for area, group in ag.iteritems():
                    # Look up area number to find max group number
                    if area_number == area:

                        # For little odds (odds between 1 and 9)
                        if int(group) in odd_one:
                            if int(group_number) in odd_one and int(group_number) <= int(group):
                                #print '%s-%s-%s valid SSN' %(area_number, group_number, serial_number)
                                return (True, True)
                            else:
                                return (False, False)

                        # For big evens (evens between 10 and 98)
                        elif int(group) in even_two:
                            ok_areas = odd_one + [a for a in even_two if a <= int(group)]
                            #print ok_areas
                            if int(group_number) in ok_areas:
                                #print '%s-%s-%s valid SSN' %(area_number, group_number, serial_number)
                                return (True, True)
                            else:
                                return (False, False)

                        # For little evens (evens between 2 and 8)
                        elif int(group) in even_three:
                            ok_areas = odd_one + even_two + [a for a in even_three if a <= int(group)]
                            #print ok_areas
                            if int(group_number) in ok_areas:
                                #print '%s-%s-%s valid SSN' %(area_number, group_number, serial_number)
                                return (True, True)
                            else:
                                return (False, False)

                        # For big odds (odds between 11 and 99)
                        elif int(group) in odd_four:
                            ok_areas = odd_one + even_two + even_three + [a for a in odd_four if a <= int(group)]
                            #print ok_areas
                            if int(group_number) in ok_areas:
                                #print '%s-%s-%s valid SSN' %(area_number, group_number, serial_number)
                                return (True, True)
                            else:
                                return (False, False)

                        else:
                            pass

            else:
                # Bad area number.
                return (False, False)

            # End of valid_ssn() function.


        for number in numbers:

            #print number

            # Remove delimiters from numbers
            clean_number = re.sub(bad, "", number)

            if clean_number not in skipped_numbers:

                #print clean_number

                # Validate CCNs
                if len(clean_number) > 9:
                    is_it = valid_ccn(clean_number)
                    if is_it[0] == True:
                        #print clean_number
                        if f not in suspect:
                            suspect.append(f)
                            print>> details, "\n\n", f, "\n", number
                        else:
                            suspect.append(f)
                            print>> details, number
                    else:
                        pass

                # Validate SSNs
                else:
                    digits = [clean_number[0:3], clean_number[3:5], clean_number[5:9]]
                    is_it = valid_ssn(digits[0], digits[1], digits[2])
                    if is_it[0] == True:
                        if f not in suspect:
                            suspect.append(f)
                            print>> details, "\n\n", f, "\n", number
                        else:
                            suspect.append(f)
                            print>> details, number
                    else:
                        pass

            else:
                fp = open(os.path.join(path_to_results, '.Find_SSNs_skip.txt'), 'a')
                print >> fp, f, clean_number, "\tSkipped Number"
                fp.close()


        details.close()

        # End of validate() function.

    current_count = 0

    if mode == 'gui':

        app = wx.PySimpleApp()
        app.MainLoop()

        # A progress dialog box (for Jeremy Sippel).
        dialog = wx.ProgressDialog ('Find_SSNs %s' %version.numb, '',
                                    maximum = len(target_file_list), parent = None,
                                    style = wx.PD_AUTO_HIDE | wx.PD_ELAPSED_TIME | wx.PD_ESTIMATED_TIME)


    for f in target_file_list:

        try:
            # Determine if the file is a PDF file
            is_pdf = False
            fd = open(f, 'rb')
            # This reads '%PDF-1.' if it is present.
            header = fd.read(7)
            fd.close()
            #print header
            if header == '%PDF-1.':
                is_pdf = True
                #print f, "is a pdf file."

            # ad hoc header skipping

            # Skip TTF's
            if ord(header[0]) == 0 and ord(header[1]) == 1 and ord(header[2]) == 0 and ord(header[3]) == 0 and ord(header[4]) == 0:
                continue

            # Skip TTC's
            if ord(header[0]) == 116 and ord(header[1]) == 116 and ord(header[2]) == 99 and ord(header[3]) == 102:
                continue
            
            # EXE's and DLL's
            if ord(header[0]) == 77 and ord(header[1]) == 90:
                continue
            
            

        except Exception, e:
            fp = open(os.path.join(path_to_results, '.Find_SSNs_exceptions.txt'), 'a')
            print >> fp, __name__, e, f, "is_pdf"
            fp.close()

        #print f

        current_count = current_count + 1

        if mode == 'gui':

            # A progress dialog box (for Jeremy Sippel).
            dialog.Update(current_count, 'Searching file %s of %s' %(current_count, len(target_file_list)))

        else:

            print "Searching file %s of %s" %(current_count, len(target_file_list))


        # Handle zip archives (zip, OOo XML and MS Office XML)
        if zipfile.is_zipfile(f):

            try:
                x = zipfile.ZipFile(f)
                zfiles = x.namelist()
                for zf in zfiles:
                    data = x.read(zf)
                    if data:
                        search = reg_exp.findall(data)
                        if search:
                            #print search
                            validate(search, f)
                x.close()

            except Exception, e:
                fp = open(os.path.join(path_to_results, '.Find_SSNs_exceptions.txt'), 'a')
                print >> fp, __name__, e, f, "is_zipfile"
                fp.close()


        # Option to handle PDF files. Off by default.
        #
        # You must have pdftotext installed and working on your system before enabling this feature.
        # pdftotext is a part of the xpdf project. It is free software. It may be downloaded here:
        # http://www.foolabs.com/xpdf/download.html
        #
        # To enable pdf searching, uncomment the following section:

##        elif is_pdf:
##            try:
##                #print f, "is a pdf file."
##
##                sfd = os.popen("pdftotext %s -" %f)
##                pdf_text = sfd.read()
##                sfd.close()
##
##                search = reg_exp.findall(pdf_text)
##                if search:
##                    validate(search, f)
##
##            except Exception, e:
##                fp = open(os.path.join(path_to_results, '.Find_SSNs_exceptions.txt'), 'a')
##                print >> fp, __name__, e, f
##                fp.close()


        # Handle most other files.
        else:

            try:
                target_file = open(f, 'rb')
                while 1:
                    data = target_file.read(256)
                    if data:
                        search = reg_exp.findall(data)
                        if search:
                            #print search
                            validate(search, f)
                    else:
                        break
                target_file.close()

            except Exception, e:
                fp = open(os.path.join(path_to_results, '.Find_SSNs_exceptions.txt'), 'a')
                print >> fp, __name__, e, f, "most_other_files"
                fp.close()

    if mode == 'gui':
        dialog.Destroy()

    uniqueSet = list(set(suspect))
    suspect_numbers = 0
    for item in uniqueSet:
        if suspect.count(item) >= threshold:
            suspect_numbers += suspect.count(item)

    # Write Report

    if report_type == 'html':

        if suspect_numbers == 0:
            print>> results, '<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 3.2//EN">'
            print>> results, "<html>"
            print>> results, "<head>"
            print>> results, "<title>'Find_SSNs %s'</title>" %(version.numb)
            print>> results, "</head>"
            print>> results, "<body>"
            print>> results, "<h1>Find_SSNs %s</h1>" %(version.numb)
            print>> results, "Host ", version.host, " | "
            print>> results, version.ip, " | "
            print>> results, tt_time, " | "
            print>> results, "Total Suspect Numbers ", suspect_numbers, " | "
            print>> results, "Threshold ", threshold, " | "
            print>> results, "Max File Size ", ms, " | "
            print>> results, "Search For ", search_for_value, "<br><br>"
            print>> results, "Search Path ", search_path, "<br><br>"
            print>> results, "<p style='color:blue'>No Sensitive Numbers Were Found During This Scan</p>"
            print>> results, "<br>End of report"
            print>> results, "</body>"
            print>> results, "</html>"
        else:
            print>> results, '<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 3.2//EN">'
            print>> results, "<html>"
            print>> results, "<head>"
            print>> results, "<title>'Find_SSNs %s'</title>" %(version.numb)
            print>> results, "<style type='text/css'>"
            print>> results, "h3:target {color: white; background: #FF0000;}"
            print>> results, "</style>"
            print>> results, '<script src="http://www.security.vt.edu/Find_SSNs/sorttable.js"></script>'
            print>> results, "</head>"
            print>> results, "<body>"
            print>> results, "<h1>Find_SSNs %s</h1>" %(version.numb)
            print>> results, "Host ", version.host, " | "
            print>> results, version.ip, " | "
            print>> results, tt_time, " | "
            print>> results, "Total Suspect Numbers ", suspect_numbers, " | "
            print>> results, "Threshold ", threshold, " | "
            print>> results, "Max File Size ", ms, " | "
            print>> results, "Search For ", search_for_value, "<br><br>"
            print>> results, "Search Path ", search_path, "<br><br>"
            print>> results, "<table border='1' summary='results' class='sortable'>"
            print>> results, "<tr>"
            print>> results, "<td align='center' bgcolor='cccccc' style='cursor:crosshair'>Suspect Number Count</td>"
            print>> results, "<td align='center' bgcolor='cccccc' style='cursor:crosshair'>File Extension</td>"
            print>> results, "<td align='center' bgcolor='cccccc' style='cursor:crosshair'>File Path</td>"
            print>> results, "</tr>"
            for item in uniqueSet:
                if suspect.count(item) >= threshold:
                    #print item
                    print>> results, "<tr><td align='center'><a href='Find_SSNs.txt' title='View the suspect numbers'>%s</a></td>" %(suspect.count(item))
                    print>> results, "<td align='center'>%s</td>" %os.path.splitext(item)[1]
                    print>> results, '<td><a href="file:///%s" title="Open the file">%s</a></td></tr>' %(item, item)
            print>> results, "</table>"
            print>> results, "<br>End of report"
            print>> results, "</body>"
            print>> results, "</html>"

    if report_type == 'csv':
        if suspect_numbers == 0:
            results.write("#,Ext,Path,Host,IP,Date,Total,Threshold,Maxsize,Search_For,Search_Path\n")
            results.write("NO,DATA,FOUND,%s,%s,%s,%s,%s,%s,%s,%s\n" %(version.host,
                                                                      version.ip, 
                                                                      tt_time,
                                                                      suspect_numbers,
                                                                      threshold,
                                                                      ms,
                                                                      search_for_value,
                                                                      search_path))
        else:
            results.write("#,Ext,Path,Host,IP,Date,Total,Threshold,Maxsize,Search_For,Search_Path\n")
           
            for item in uniqueSet:
                if suspect.count(item) >= threshold:
                    #print item
                    results.write("%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s\n" %(suspect.count(item),
                                                                      os.path.splitext(item)[1],
                                                                      item,
                                                                      version.host,
                                                                      version.ip,   
                                                                      tt_time,
                                                                      suspect_numbers,
                                                                      threshold,
                                                                      ms,
                                                                      search_for_value,
                                                                      search_path))



                    
    results.close()
    
